
"""
Code implemented for individual regressions for Master's Thesis research,
Master of Science in Business, Major in Finance.
Authors: Lars Fredrik Osen and Jacob-Sterling Deehr Johannessen

"""

import pandas as pd
from statsmodels.regression.linear_model import OLS
from statsmodels.tools.tools import add_constant
from itertools import product

import pandas as pd
import numpy as np
from pandas.tseries.offsets import *
from scipy import stats
import datetime as dt
from pandas.tseries.offsets import MonthBegin
import matplotlib.pyplot as plt
# Load data
SPX = pd.read_excel('/Users/larsfredrikosen/Desktop/MSC_DATA_THESIS/Data/SPX.xlsx')
SECTOR = pd.read_excel('/Users/larsfredrikosen/Desktop/MSC_DATA_THESIS/Data/SECTOR.xlsx', index_col = 0)
SPX_index = pd.read_excel('/Users/larsfredrikosen/Desktop/MSC_DATA_THESIS/Data/SPX_index.xlsx', index_col=0)
M2 = pd.read_excel('/Users/larsfredrikosen/Desktop/MSC_DATA_THESIS/Data/M2.xlsx', index_col = 0)
GOLD = pd.read_excel('/Users/larsfredrikosen/Desktop/MSC_DATA_THESIS/Data/GOLD.xlsx', index_col = 0)
RF = pd.read_excel('/Users/larsfredrikosen/Desktop/MSC_DATA_THESIS/Data/RF.xlsx', index_col = 0)
CPI = pd.read_excel('/Users/larsfredrikosen/Desktop/MSC_DATA_THESIS/Data/CPIAUCSL.xlsx', index_col = 0)


#### Calculating returns #### 

# GOLD Monthly Returns
GOLD_M_ret = GOLD.pct_change().resample('M').agg(lambda x: (x + 1).prod() - 1)

# CPIAUCSL % Change
CPI_M_ret = CPI.pct_change().resample('M').agg(lambda x: (x + 1).prod() - 1)

# M2 % Change
M2_M_ret = M2.pct_change().resample('M').agg(lambda x: (x + 1).prod() - 1)



SPX['Date'] = pd.to_datetime(SPX['Date'])

SPX.set_index('Date', inplace=True)

SPX_M = SPX.resample('M').last()

SPX_M = SPX_M.reset_index()


stock_prices_melted = pd.melt(SPX_M, id_vars=['Date'], var_name='stock', value_name='price')

stock_prices_melted['Date'] = pd.to_datetime(stock_prices_melted['Date'])


# Create a dictionary mapping stocks to ID_nos
stock_to_ID_no = dict(zip(stock_prices_melted['stock'].unique(), range(1, len(stock_prices_melted['stock'].unique())+1)))

# Map stocks to ID_nos and add as a new column to the DataFrame
stock_prices_melted['ID_no'] = pd.factorize(stock_prices_melted['stock'], sort=True)[0] + 1


# Sort DataFrame by date and ID_no
stock_prices_melted = stock_prices_melted.sort_values(['Date', 'ID_no'])

# Group DataFrame by ID_no and calculate monthly returns
stock_prices_melted['ret'] = stock_prices_melted.groupby('ID_no')['price'].pct_change()


## Rename stock prices_melted to SPX_LP_M
SPX_LP_m = stock_prices_melted

SPX_LP_m.tail()

# Line up date to be end of month
SPX_LP_m['Date']=pd.to_datetime(SPX_LP_m['Date'])


SPX_LP_m.tail()
#######################################################
# Create Momentum Portfolio                           #   
# Measures Based on Past (J) Month Compounded Returns #
#######################################################

def calculate_momentum(J, K, SPX_LP_m):

    _tmp_SPX_LP = SPX_LP_m[['ID_no', 'Date', 'ret']].sort_values(['ID_no', 'Date']).set_index('Date')
    _tmp_SPX_LP['ret'] = _tmp_SPX_LP['ret'].fillna(0)
    _tmp_SPX_LP['logret'] = np.log(1 + _tmp_SPX_LP['ret'])
    umd = _tmp_SPX_LP.groupby(['ID_no'])['logret'].rolling(J, min_periods=J).sum().reset_index()
    umd['cumret'] = np.exp(umd['logret']) - 1
    umd[['cumret', 'logret']] = umd[['cumret', 'logret']].fillna(0)
    umd[['cumret', 'logret']] = umd.groupby('ID_no')[['cumret', 'logret']].shift(K)
    umd = umd.dropna()

    umd=umd.dropna(axis=0, subset=['cumret'])
    umd['momr'] = umd.groupby('Date')['cumret'].transform(lambda x: pd.qcut(x[x != 0], 10, labels=False, duplicates='drop') if not x[x != 0].empty else pd.Series(np.nan, index=x.index))

    umd['form_date'] = umd['Date']
    umd['medate'] = umd['Date']+MonthEnd(0)
    umd['hdate1']=umd['medate']+MonthBegin(1)
    umd['hdate2']=umd['medate']+MonthEnd(K)
    umd = umd[['ID_no', 'form_date','momr','hdate1','hdate2']]

    _tmp_ret = SPX_LP_m[['ID_no','Date','ret']]
    port = pd.merge(_tmp_ret, umd, on=['ID_no'], how='inner')
    port = port[(port['hdate1']<=port['Date']) & (port['Date']<=port['hdate2'])]

    umd2 = port.sort_values(by=['Date','momr','form_date','ID_no']).drop_duplicates()
    umd3 = umd2.groupby(['Date','momr','form_date'])['ret'].mean().reset_index()

    start_yr = umd3['Date'].dt.year.min()+2
    umd3 = umd3[umd3['Date'].dt.year>=start_yr]
    umd3 = umd3.sort_values(by=['Date','momr'])

    ewret = umd3.groupby(['Date','momr'])['ret'].mean().reset_index()
    ewstd = umd3.groupby(['Date','momr'])['ret'].std().reset_index()
    ewret = ewret.rename(columns={'ret':'ewret'})
    ewstd = ewstd.rename(columns={'ret':'ewretstd'})
    ewretdat = pd.merge(ewret, ewstd, on=['Date','momr'], how='inner')
    ewretdat = ewretdat.sort_values(by=['momr'])

    ewretdat2 = ewretdat.pivot(index='Date', columns='momr', values='ewret')
    ewretdat2 = ewretdat2.add_prefix('port')
    ewretdat2 = ewretdat2.rename(columns={'port0.0':'losers','port9.0':'winners'})
    ewretdat2['long_short'] = ewretdat2['winners'] - ewretdat2['losers']

    ewretdat3 = ewretdat2
    ewretdat3['1+losers']=1+ewretdat3['losers']
    ewretdat3['1+winners']=1+ewretdat3['winners']
    ewretdat3['1+ls'] = 1+ewretdat3['long_short']

    ewretdat3['cumret_winners']=ewretdat3['1+winners'].cumprod()-1
    ewretdat3['cumret_losers']=ewretdat3['1+losers'].cumprod()-1
    ewretdat3['cumret_long_short']=ewretdat3['1+ls'].cumprod()-1

    mom_mean = ewretdat3[['winners', 'losers', 'long_short']].mean().to_frame()
    mom_mean = mom_mean.rename(columns={0:'mean'}).reset_index()

    t_losers = pd.Series(stats.ttest_1samp(ewretdat3['losers'],0.0)).to_frame().T
    t_winners = pd.Series(stats.ttest_1samp(ewretdat3['winners'],0.0)).to_frame().T
    t_long_short = pd.Series(stats.ttest_1samp(ewretdat3['long_short'],0.0)).to_frame().T

    t_losers['momr']='losers'
    t_winners['momr']='winners'
    t_long_short['momr']='long_short'

    t_output =pd.concat([t_losers, t_winners, t_long_short])\
        .rename(columns={0:'t-stat', 1:'p-value'})

    mom_output = pd.merge(mom_mean, t_output, on=['momr'], how='inner')
    
    return ewretdat3, mom_output

# Values for J and K
Js = [1, 3, 6, 9, 12]
Ks = [1, 3, 6, 9, 12]

# Create DataFrames
long_short_all = pd.DataFrame()
long_all = pd.DataFrame()
short_all = pd.DataFrame()

mom_output = pd.DataFrame()

# For every combination of J and K
for J in Js:
    for K in Ks:
        # Calculate the long-short series for JxK
        ewretdat3, mom_output = calculate_momentum(J, K, SPX_LP_m)

        # Extract strategies from ewretdat3
        long_short = ewretdat3['long_short']
        winners = ewretdat3['winners']
        losers = -ewretdat3['losers']

        # Join it into DF, with a new column name for each JxK combination
        long_short_all = pd.concat([long_short_all, long_short.rename(f'long_short_J{J}_K{K}')], axis=1)
        long_all = pd.concat([long_all, winners.rename(f'winners_J{J}_K{K}')], axis=1)
        short_all = pd.concat([short_all, losers.rename(f'losers_J{J}_K{K}')], axis=1)
        

## SPX INDEX CUMULAITIVE RETURNS 
spx_index_m = SPX_index.resample('M').last()
spx_index_returns = spx_index_m.pct_change()
cumulativ_returns_spx_index = (1+ spx_index_returns).cumprod()-1

start_date = pd.to_datetime('1983-01-01')
spx_index_returns_filtered = spx_index_returns[spx_index_returns.index >= start_date]
cumulative_returns_spx_index_filtered = (1 + spx_index_returns_filtered).cumprod()-1


### Returns Output ###

timestamp = pd.Timestamp('2008-11-01')

# Split the data into three periods
pre_nov_2008_long_short = long_short_all[long_short_all.index < timestamp]
post_nov_2008_long_short = long_short_all[long_short_all.index >= timestamp]

pre_nov_2008_long = long_all[long_all.index < timestamp]
post_nov_2008_long = long_all[long_all.index >= timestamp]

pre_nov_2008_short = short_all[short_all.index < timestamp]
post_nov_2008_short = short_all[short_all.index >= timestamp]


# Save the data to Excel file
with pd.ExcelWriter('Momentum_returns_results.xlsx') as writer:  
    # Total Period
    long_short_all.to_excel(writer, sheet_name='Total_Period_Long_Short')
    long_all.to_excel(writer, sheet_name='Total_Period_Long')
    short_all.to_excel(writer, sheet_name='Total_Period_Short')

    # Pre November 2008
    pre_nov_2008_long_short.to_excel(writer, sheet_name='Pre_Nov_2008_Long_Short')
    pre_nov_2008_long.to_excel(writer, sheet_name='Pre_Nov_2008_Long')
    pre_nov_2008_short.to_excel(writer, sheet_name='Pre_Nov_2008_Short')

    # Post November 2008
    post_nov_2008_long_short.to_excel(writer, sheet_name='Post_Nov_2008_Long_Short')
    post_nov_2008_long.to_excel(writer, sheet_name='Post_Nov_2008_Long')
    post_nov_2008_short.to_excel(writer, sheet_name='Post_Nov_2008_Short')


# Calculate cumulative returns and mean returns
cumulative_returns = {
    'long_short_all': long_short_all.sum(),
    'long_all': long_all.sum(),
    'short_all': short_all.sum(),
    'pre_nov_2008_long_short': pre_nov_2008_long_short.sum(),
    'pre_nov_2008_long': pre_nov_2008_long.sum(),
    'pre_nov_2008_short': pre_nov_2008_short.sum(),
    'post_nov_2008_long_short': post_nov_2008_long_short.sum(),
    'post_nov_2008_long': post_nov_2008_long.sum(),
    'post_nov_2008_short': post_nov_2008_short.sum()
}

mean_returns = {
    'long_short_all': long_short_all.mean(),
    'long_all': long_all.mean(),
    'short_all': short_all.mean(),
    'pre_nov_2008_long_short': pre_nov_2008_long_short.mean(),
    'pre_nov_2008_long': pre_nov_2008_long.mean(),
    'pre_nov_2008_short': pre_nov_2008_short.mean(),
    'post_nov_2008_long_short': post_nov_2008_long_short.mean(),
    'post_nov_2008_long': post_nov_2008_long.mean(),
    'post_nov_2008_short': post_nov_2008_short.mean()
}


######################
# Regression Outputs #
######################

## Combined regressions

import statsmodels.api as sm

# Set date as index for each DF
RF.index = pd.to_datetime(RF.index)
GOLD.index = pd.to_datetime(GOLD.index)
CPI.index = pd.to_datetime(CPI.index)
M2.index = pd.to_datetime(M2.index)
long_short_all.index = pd.to_datetime(long_short_all.index)

# Resample the data to EOM
RF_month_end = RF.resample('M').last()
GOLD_month_end = GOLD.resample('M').last()
CPI_month_end = CPI.resample('M').last()
M2_month_end = M2.resample('M').last()
long_short_all_month_end = long_short_all.resample('M').last()

# DF to store the results
results_df = pd.DataFrame()

# Iterate over all columns of long_short_all
for col in long_short_all_month_end.columns:
    # Merging all the dataframes
    data = pd.concat([long_short_all_month_end[col], RF_month_end, CPI_month_end, M2_month_end, GOLD_month_end], axis=1, join='inner')
    data.columns = ['momentum', 'RF', 'CPI', 'M2', 'GOLD']

    # Dropping rows with any NaN values
    data = data.dropna()

    # Splitting data into two based on the date (before and after November 2008)
    data_pre_2008 = data[data.index < '2008-11']
    data_post_2008 = data[data.index >= '2008-11']

    # Running the regression for each period
    for data_period, period_name in zip([data_pre_2008, data_post_2008], ['pre_2008', 'post_2008']):
        y = data_period['momentum']
        X = sm.add_constant(data_period[['RF', 'CPI', 'M2', 'GOLD']])
        model = sm.OLS(y, X)
        results = model.fit()

        # Adding the results to the results DF
        results_df = results_df.append(pd.Series({
            'column': col,
            'period': period_name,
            'R_squared': results.rsquared,
            'const_coef': results.params['const'],
            'gold_coef': results.params['GOLD'],
            'cpi_coef': results.params['CPI'],
            'm2_coef': results.params['M2'],
            'rf_coef': results.params['RF'],
            'const_pvalue': results.pvalues['const'],
            'gold_pvalue': results.pvalues['GOLD'],
            'cpi_pvalue': results.pvalues['CPI'],
            'm2_pvalue': results.pvalues['M2'],
            'rf_pvalue': results.pvalues['RF']
        }, name=f'{col}_{period_name}'))

# Write the results to an Excel file
results_df.to_excel('momentum_regression_results_0.xlsx')


## Individual regressions 

import pandas as pd
import statsmodels.api as sm
import openpyxl


# Resample the data to end-of-month
RF_month_end = RF.resample('M').last()
GOLD_month_end = GOLD.resample('M').last()
M2_month_end = M2.resample('M').last()
CPI_month_end = CPI.resample('M').last()
long_short_all_month_end = short_all.resample('M').last()

# Create a new Excel writer object
writer = pd.ExcelWriter('momentum_regression_results_1.xlsx', engine='openpyxl')

# For each of the variables RF, GOLD, and M2
for var, var_name in zip([RF_month_end, GOLD_month_end, M2_month_end, CPI_month_end], ['RF', 'GOLD', 'M2', 'CPI']):
    # Initialize an empty DataFrame to store the results
    results_df = pd.DataFrame()

    # Iterate over all columns of long_short_all
    for col in long_short_all_month_end.columns:
        # Merging the series
        data = pd.concat([long_short_all_month_end[col], var], axis=1, join='inner')
        data.columns = ['momentum', var_name]

        # Dropping rows with any NaN values
        data = data.dropna()

        # Splitting data into two based on the date (before and after November 2008)
        data_pre_2008 = data[data.index < '2008-11']
        data_post_2008 = data[data.index >= '2008-11']

        # Running the regression for each period
        for data_period, period_name in zip([data_pre_2008, data_post_2008], ['pre_2008', 'post_2008']):
            y = data_period['momentum']
            X = sm.add_constant(data_period[[var_name]])
            model = sm.OLS(y, X)
            results = model.fit()

            # Adding the results to the results DataFrame
            results_df = results_df.append(pd.Series({
                'column': col,
                'period': period_name,
                'R_squared': results.rsquared,
                'const_coef': results.params['const'],
                f'{var_name}_coef': results.params[var_name],
                'const_tvalue': results.tvalues['const'],
                f'{var_name}_tvalue': results.tvalues[var_name],
                'const_pvalue': results.pvalues['const'],
                f'{var_name}_pvalue': results.pvalues[var_name]
            }, name=f'{col}_{period_name}'))

    # Write the results to the appropriate sheet in the Excel file
    results_df.to_excel(writer, sheet_name=var_name)

# Save the Excel file
writer.save()
